/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "ftp_main.h"

static int child_setup(struct proxyinfo* pi);

struct config_directives_list main_configuration_directives_array[] = {
	{"access",		TAG_ALL, "deny"   , EM, WSP },
	{"account",		TAG_CONNECTED, (char*) 0, EM, FL },
	{"activeportrangeclient",	TAG_ALL, (char*) 0, EM, FL },
	{"activeportrangeserver",	TAG_ALL, (char*) 0, EM, FL },
	{"activeportrange",	TAG_ALL, (char*) 0, EM, FL },
	{"allowforeignaddress",	TAG_ALL, "no", EM, WSP },
	{"allowforeigndataconnect",	TAG_ALL, "no", EM, WSP },
	{"allowreservedports",	TAG_ALL, "no", EM, WSP },
	{"cachemaxsize",	TAG_ALL, "unlimited", EM, WSP },
	{"cacheminsize",	TAG_ALL,       "0", EM, WSP },
	{"cacheprefix",		TAG_ALL, (char*) 0, EM, WSP },
	{"cache",		TAG_ALL,  "off"   , EM, WSP },
	{"changerootdir",	TAG_STARTUP, (char*) 0, EM, WSP },
	{"changeroot",		TAG_STARTUP, "never", EM, WSP },
	{"cmdlogfile-specs",	TAG_ALL, (char*) 0, EM, FL },
	{"cmdlogfile-style",	TAG_ALL, (char*) 0, EM, FL },
	{"cmdlogfile",		TAG_ALL, (char*) 0, EM, FL },
	{"commandtimeout",	TAG_ALL, "300", EM, WSP },
	{"connectionlogdir-fileprefix",	TAG_ALL, (char*) 0, EM, WSP },
	{"connectionlogdir-filesuffix",	TAG_ALL, (char*) 0, EM, WSP },
	{"connectionlogdir-specs",	TAG_ALL, (char*) 0, EM, FL },
	{"connectionlogdir-style",	TAG_ALL, (char*) 0, EM, FL },
	{"connectionlogdir",	TAG_ALL, (char*) 0, EM, WSP },
	{"controlserveraddress",TAG_ALL, (char*) 0, EM, WSP },
	{"dataclientaddress",	TAG_ALL, (char*) 0, EM, WSP },
	{"dataserveraddress",	TAG_ALL, (char*) 0, EM, WSP },
	{"debuglevel",		TAG_ALL, "7"      , EM, WSP },
	{"defaultmode",		TAG_ALL, "asclient", EM, WSP },
	{"displayclientaddress",TAG_ALL, (char*) 0, EM, WSP },
	{"displayserveraddress",TAG_ALL, (char*) 0, EM, WSP },
	{"dnslookups",		TAG_ALL, "yes", EM, WSP },
	{"dropprivileges",	TAG_STARTUP, "start", EM, WSP },
	{"failedlogins",	TAG_ALL,       "3", EM, WSP },
	{"filterorderdownstream",	TAG_STARTUP, (char*) 0, EM, FL },
	{"filterorderupstream",	TAG_STARTUP, (char*) 0, EM, FL },
	{"forwardlookups",	TAG_ALL, "yes", EM, WSP },
	{"forward",		TAG_ALL, (char*) 0, EM, FL },
	{"forwardcmds",		TAG_ALL, (char*) 0, EM, FL },
	{"getinternalip",	TAG_CONNECTED, "udp", EM, FL },
	{"hostcachetimeout",	TAG_ALL, "28800", EM, WSP },
	{"initialsyst",		TAG_ALL, "yes", EM, WSP },
	{"libwrapservicenames",	TAG_STARTUP, "ftp-gw jftpgw", EM, FL },
	{"limit",		TAG_CONNECTED, (char*) 0, EM, WSP },
	{"listen",		TAG_STARTUP, "0.0.0.0:2370", EM, FL },
	{"logfile",		TAG_ALL, "/var/log/jftpgw.log", EM, WSP },
	{"loginstyle",		TAG_CONNECTED, "1", EM, WSP },
	{"logintime",		TAG_TO | TAG_CONNECTED, "user", EM, WSP },
	{"logstyle",		TAG_ALL, "files"  , EM, WSP },
	{"passiveportrangeclient",	TAG_ALL, (char*) 0, EM, FL },
	{"passiveportrangeserver",	TAG_ALL, (char*) 0, EM, FL },
	{"passiveportrange",	TAG_ALL, (char*) 0, EM, FL },
	{"pidfile",		TAG_STARTUP, "/var/run/jftpgw.pid", EM, WSP },
	{"protocolviolations",	TAG_ALL, "10", EM, WSP },
	{"reverselookups",	TAG_ALL, "yes", EM, WSP },
	{"runasgroup",		TAG_STARTUP, (char*) 0, EM, WSP },
	{"runasuser",		TAG_STARTUP, (char*) 0, EM, WSP },
	{"sendcapa",		TAG_ALL, "yes", EM, WSP },
	{"serverport",		TAG_ALL, "21", EM, WSP },
	{"strictasciiconversion",	TAG_ALL, "on", EM, WSP },
	{"syslogfacility",	TAG_ALL, "daemon", EM, WSP },
	{"syslogcmdlogpriority",TAG_ALL, "warning", EM, WSP },
	{"syslogdebugpriority",	TAG_ALL, "warning", EM, WSP },
	{"throughput",		TAG_ALL, (char*) 0, EM, WSP },
	{"throughput-up",	TAG_ALL, (char*) 0, EM, WSP },
	{"throughput-down",	TAG_ALL, (char*) 0, EM, WSP },
	{"transfertimeout",	TAG_ALL, "300", EM, WSP },
	{"transparent-forward-include-port", TAG_CONNECTED, "on", EM, WSP },
	{"transparent-forward",	TAG_CONNECTED, (char*) 0, EM, WSP },
	{"transparent-proxy",	TAG_CONNECTED, "off", EM, WSP },
	{"udpport",		TAG_CONNECTED, "2370", EM, FL },
	{"welcomeline",		TAG_CONNECTED,
			"FTP proxy (v"PACKAGE_VERSION") ready", EM, FL },
						/* 8 hours */
	{ (char*) 0,                          0, (char*) 0, 0, 0 }
};

struct config_namespace_directives_list main_configuration_directives = {
	"main",       /* namespace name */
	main_configuration_directives_array
};

struct config_option_values_list main_configuration_options_array[] = {
	{"access",              {"allow", "deny", TERM} },
	{"allowforeignaddress", { TRUEFALSE, TERM } },
	{"allowreservedports",  { TRUEFALSE, TERM } },
	{"cache",               { TRUEFALSE, TERM } },
	/*                           <---- better                          */
	{"changeroot",          {"start", "startsetup", "connect",
				  "connectsetup", "loggedin", "never", TERM} },
	{"defaultmode",         {"asclient", "active", "passive", TERM} },
	{"dnslookups",          { TRUEFALSE, TERM } },
	/*                           <---- better                          */
	{"dropprivileges",      {"start", "startsetup", "connect",
				  "connectsetup", "loggedin", "never", TERM} },
	{"forwardlookups",      { TRUEFALSE, TERM } },
	{"getinternalip",       {"udp", "icmp", "configuration", TERM} },
	{"logintime",           {"connect", "user", "pass", TERM} },
	{"logstyle",            {"files", "syslog", TERM} },
	{"reverselookups",      { TRUEFALSE, TERM } },
	{"transparent-proxy",	{ TRUEFALSE, TERM } },
	{"sendcapa",		{ TRUEFALSE, TERM } },
	{"syslogfacility", {
#ifdef HAVE_LOG_FACILITY_LOG_AUTH
				"auth",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_AUTHPRIV
				"authpriv",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_CRON
				"cron",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_DAEMON
				"daemon",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_FTP
				"ftp",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_KERN
				"kern",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL0
				"local0",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL1
				"local1",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL2
				"local2",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL3
				"local3",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL4
				"local4",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL5
				"local5",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL6
				"local6",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL7
				"local7",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LPR
				"lpr",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_MAIL
				"mail",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_NEWS
				"news",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_SYSLOG
				"syslog",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_USER
				"user",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_UUCP
				"uucp",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_CONSOLE
				"console",
#endif
#ifdef HAVE_LOG_FACILITY_LOG_SECURITY
				"security",
#endif
				TERM } },
	{ TERM, { TERM } }
};

struct config_namespace_options_list main_configuration_options = {
	"main",
	main_configuration_options_array
};

int ftp_add_configuration() {
	config_configuration_list_add(&main_configuration_directives);
	config_options_list_add(&main_configuration_options);
	return 0;
}


static
int child_setup(struct proxyinfo *pi) {
	int cs = pi->si->connection.clientsocket;

	/* Init proxy components - in contrast to the modules which are already
	 * init()ed because they had to add their configuration data before
	 * reading the configuration file */
	capa_init(pi);
	transfermode_init();
	ftp_filter_init();

	/* now that the filters are initialized - do binary transfers by
	 * default */
	pi->ci->data_connection.asciiconversion.serverlisting = 0;
	pi->ci->data_connection.asciiconversion.type = 'a';
        filter_activate("asciiconversion", FILTER_TYPE_DOWNSTREAM);
	filter_activate("asciiconversion", FILTER_TYPE_UPSTREAM);

	if (stage_action("connectsetup", pi) < 0) {
		say(cs, "421 Error setting up (see logfile)\r\n");
		return -1;
	}

	return 0;
}

int ftp_allocate(struct proxyinfo* pi) {
	struct clientinfo* ci = malloc(sizeof(struct clientinfo));
	struct serverinfo* si = pi->si;

	enough_mem(ci);
	memset(ci, 0, sizeof(struct clientinfo));

	pi->ci = ci;
	ci->service = "ftp";

	/* initialize a few values */
	ci->data_connection.transfermode.servermode = UNSPEC;
	ci->data_connection.transfermode.clientmode = UNSPEC;
	ci->data_connection.portcmd = (char*) 0;
	ci->data_connection.dataclientsock = -1;
	ci->data_connection.dataserversock = -1;
	ci->data_connection.dataport
		= socketinfo_get_local_port(si->connection.clientsocket) - 1;
	jlog(9, "dataport is set to %d", ci->data_connection.dataport);
	/* No login done yet */
	pi->ci->login.stage = LOGIN_ST_NOT_CONNECTED;

	return 0;
}

int ftp_deallocate(struct proxyinfo* pi) {
	struct clientinfo* ci = pi->ci;
	struct serverinfo* si = pi->si;

	/* XXX where should it go? should the init be in allocate? */
	transfermode_deinit();

	if (!ci) {
		/* if we didn't call proto_accept() (because we are the master
		 * process), there is no deallocation */
		return 0;
	}

	/* maybe these variables are not yet allocated */
	/* XXX std_reset */
	if (ci->user) {
		free(ci->user);
	}
	if (ci->user_login) {
		free(ci->user_login);
	}
	if (ci->pass) {
		free(ci->pass);
	}
	if (ci->destination) {
		free(ci->destination);
	}
	if (ci->anon_user_pass) {
		free_and_zero_char(&ci->anon_user_pass);
	}
	if (ci->forward.login) {
		free_and_zero_char(&ci->forward.login);
	}
	if (ci->forward.original.user) {
		free_and_zero_char(&ci->forward.original.user);
	}
	if (ci->forward.original.destination) {
		free_and_zero_char(&ci->forward.original.destination);
	}

	destroy_active_portrange();
	destroy_passive_portrange();

	close(si->connection.clientsocket);
	close(si->connection.serversocket);

	free(ci);
	pi->ci = (struct clientinfo*) 0;
	return 0;
}

int ftp_reject(struct proxyinfo* pi) {
	int cs = pi->si->connection.clientsocket;
	say(cs, "500 Too many connections, sorry\r\n");
	close(cs);
	return 0;
}

static
int ftp_install_standard_commandhandlers(struct proxyinfo* pi) {
	jlog(7, "Installing standard commandhandlers");
	std_cmds_init();
	jlog(7, "Loading commandhandlers from modules");
	event_fire(EVENT_FTP_CMDHANDLER_INIT, pi);
	return 0;
}

int ftp_connect(struct proxyinfo* pi) {
	int cs = pi->si->connection.clientsocket;

	if (stage_action("connect", pi) < 0) {
		say(cs, "421 Error setting up (see logfile)\r\n");
		return -1;
	}
	return 0;
}

int ftp_accept(struct proxyinfo* pi) {
	int ret;
	char* msg;
	struct messagebuffer* msgbuf;
	int cs;
	int code;

	ftp_allocate(pi);

	if (child_setup(pi) < 0) {
		return -1;
	}

	/* install the right login protocol handler */
	ret = config_get_ioption("loginstyle", -1);
	/* XXX make proper error message, use constants */
	ASSUME(ret >= 0);
	ASSUME(ret <= 9);
	fw_install_handler(ret);
	event_fire(EVENT_FTP_LOGINTYPE_INIT, pi);

	/* as soon as we are logged in, we change the handler and replace the
	 * login handlers with the FTP command handlers */
	event_register(EVENT_FTP_LOGGEDIN, ftp_install_standard_commandhandlers,
			"Change from login handlers to FTP command handlers");

	msgbuf = (struct messagebuffer*) malloc(sizeof(struct messagebuffer));
	message_buffer_initialize(msgbuf);

	/* this function is run before the first command is read, hence it's
	 * outside of the commandhandler */
	ret = cmds_after_client_connect(pi, msgbuf);

	/* and we have to care for the messages for ourself */
	cs = pi->si->connection.clientsocket;
	msg = message_buffer_getmsg_m(msgbuf, cs, &code);
	if (msg) {
		say(cs, msg);
		jlog(8, "Sending message to client: %s", msg);
		free(msg);
	}
	message_buffer_reset(msgbuf);

	/* Now we call the commandhandler to handle user commands */
	ret = handle_cmds(pi, msgbuf);

	jlog(8, "Exited from handle_cmds");
	message_buffer_reset(msgbuf);
	free(msgbuf);

	return ret;
}

int ftp_read_config_before(struct proxyinfo* pi) {
	destroy_active_portrange();
	destroy_passive_portrange();
	return 0;
}

int ftp_read_config_after(struct proxyinfo* pi) {

	return 0;
}

